@@ -3,8 +3,8 @@ Provides Dataset objects for datasets in this benchmark, and various functions f
 working with the datasets or collections of datasets.
 """
 
-from .cifar import TabularCIFAR10Dataset
 from .dataset import *
+from .collection import *
 
 __all__ = [
     # functions
@@ -16,6 +16,7 @@ __all__ = [
     # classes
     'OpenTabularDataset',
     'DatasetFormatError',
+    'DatasetCollection',
     
     # data/variables/constants
     'metadata',
@@ -24,6 +25,14 @@ __all__ = [
 
 
 def list_datasets():
-    """List the non-image tabular datasets available."""
+    """
+    List the tabular datasets available.
+    
+    Returns
+    -------
+    sequence of str
+        Sequence of names of all datasets in the benchmark
+    """
+    
     return list(metadata.keys())
 
